@@ -9,7 +9,7 @@ use DynaLoader();
 use Carp ();
 @ISA = qw(DynaLoader);
 
-$VERSION = 4.001;
+$VERSION = '4.002';
 
 bootstrap DBD::mysql $VERSION;
 
@@ -98,6 +98,7 @@ sub AUTOLOAD {
 package DBD::mysql::dr; # ====== DRIVER ======
 use strict;
 use DBI qw(:sql_types);
+use DBI::Const::GetInfoType;
 
 sub connect {
     my($drh, $dsn, $username, $password, $attrhash) = @_;
@@ -252,38 +253,113 @@ sub _SelectDB ($$) {
     die "_SelectDB is removed from this module; use DBI->connect instead.";
 }
 
-{
-    my $names = ['TABLE_CAT', 'TABLE_SCHEM', 'TABLE_NAME',
-		 'TABLE_TYPE', 'REMARKS'];
+sub table_info ($) {
+    my ($dbh, $catalog, $schema, $table, $type, $attr) = @_;
+    my @names = qw(TABLE_CAT TABLE_SCHEM TABLE_NAME TABLE_TYPE REMARKS);
+    my @rows;
+
+    my $sponge = DBI->connect("DBI:Sponge:", '','')
+	or return $dbh->DBI::set_err($DBI::err, "DBI::Sponge: $DBI::errstr");
+
+    # Return the list of catalogs
+    if (defined $catalog && $catalog eq "%" &&
+        (!defined($schema) || $schema eq "") &&
+        (!defined($table) || $table eq ""))
+    {
+	@rows = (); # Empty, because MySQL doesn't support catalogs (yet)
+    }
+
+    # Return the list of schemas
+    elsif (defined $schema && $schema eq "%" &&
+           (!defined($catalog) || $catalog eq "") &&
+           (!defined($table) || $table eq ""))
+    {
+	my $sth = $dbh->prepare("SHOW DATABASES")
+	    or return undef;
+
+	$sth->execute()
+	    or return DBI::set_err($dbh, $sth->err(), $sth->errstr());
 
-    sub table_info ($) {
-	my $dbh = shift;
-	my $sth = $dbh->prepare("SHOW TABLES");
-	return undef unless $sth;
-	if (!$sth->execute()) {
-	  return DBI::set_err($dbh, $sth->err(), $sth->errstr());
-        }
-	my @tables;
 	while (my $ref = $sth->fetchrow_arrayref()) {
-	  push(@tables, [ undef, undef, $ref->[0], 'TABLE', undef ]);
-        }
-	my $dbh2;
-	if (!($dbh2 = $dbh->{'~dbd_driver~_sponge_dbh'})) {
-	    $dbh2 = $dbh->{'~dbd_driver~_sponge_dbh'} =
-		DBI->connect("DBI:Sponge:");
-	    if (!$dbh2) {
-	        DBI::set_err($dbh, 1, $DBI::errstr);
-		return undef;
+	    push(@rows, [ undef, $ref->[0], undef, undef, undef ]);
+	}
+    }
+
+    # Return the list of table types
+    elsif (defined $type && $type eq "%" &&
+           (!defined($catalog) || $catalog eq "") &&
+           (!defined($schema) || $schema eq "") &&
+           (!defined($table) || $table eq ""))
+    {
+	@rows = (
+		 [ undef, undef, undef, "TABLE", undef ],
+		 [ undef, undef, undef, "VIEW",  undef ],
+		);
+    }
+
+    # Special case: a catalog other than undef, "", or "%"
+    elsif (defined $catalog && $catalog ne "" && $catalog ne "%") {
+	@rows = (); # Nothing, because MySQL doesn't support catalogs yet.
+    }
+
+    # Uh oh, we actually have a meaty table_info call. Work is required!
+    else {
+	my @schemas;
+
+	# If no table was specified, we want them all
+	$table ||= "%";
+
+	# If something was given for the schema, we need to expand it to
+	# a list of schemas, since it may be a wildcard.
+	if (defined $schema && $schema ne "") {
+	    my $sth = $dbh->prepare("SHOW DATABASES LIKE " .
+				    $dbh->quote($schema))
+		or return undef;
+	    $sth->execute()
+		or return DBI::set_err($dbh, $sth->err(), $sth->errstr());
+
+	    while (my $ref = $sth->fetchrow_arrayref()) {
+		push @schemas, $ref->[0];
+	    }
+	# Otherwise we want the current database
+	} else {
+	    push @schemas, $dbh->selectrow_array("SELECT DATABASE()");
+	}
+
+	# Figure out which table types are desired
+	my ($want_tables, $want_views);
+	if (defined $type && $type ne "") {
+	    $want_tables = ($type =~ m/table/i);
+	    $want_views  = ($type =~ m/view/i);
+	} else {
+	    $want_tables = $want_views = 1;
+	}
+
+	foreach my $database (@schemas) {
+	    my $sth = $dbh->prepare("SHOW /*!50002 FULL*/ TABLES FROM " .
+				    $dbh->quote_identifier($database) .
+				    " LIKE " .  $dbh->quote($table))
+		or return undef;
+	    $sth->execute() or
+		return DBI::set_err($dbh, $sth->err(), $sth->errstr());
+
+	    while (my $ref = $sth->fetchrow_arrayref()) {
+		my $type = (defined $ref->[1] &&
+		            $ref->[1] =~ /view/i) ? 'VIEW' : 'TABLE';
+		next if $type eq 'TABLE' && not $want_tables;
+		next if $type eq 'VIEW'  && not $want_views;
+		push @rows, [ undef, $database, $ref->[0], $type, undef ];
 	    }
 	}
-	my $sth2 = $dbh2->prepare("SHOW TABLES", { 'rows' => \@tables,
-						   'NAME' => $names,
-						   'NUM_OF_FIELDS' => 5 });
-	if (!$sth2) {
-	    DBI::set_err($sth2, $dbh2->err(), $dbh2->errstr());
-	}
-	$sth2;
     }
+
+    my $sth = $sponge->prepare("table_info", {
+	rows => \@rows,
+	NUM_OF_FIELDS => scalar @names,
+	NAME => \@names,
+    }) or return $dbh->DBI::set_err($sponge->err(), $sponge->errstr());
+
+    return $sth;
 }
 
 sub _ListTables {
@@ -297,8 +373,10 @@ sub _ListTables {
 
 sub column_info {
     my ($dbh, $catalog, $schema, $table, $column) = @_;
-    return $dbh->set_err(1, "column_info doesn't support column selection")
-	if $column ne "%";
+    # ODBC allows a NULL to mean all columns, so we'll accept undef
+    $column = '%' unless defined $column;
+
+    my $ER_NO_SUCH_TABLE= 1146;
 
     my $table_id = $dbh->quote_identifier($catalog, $schema, $table);
 
@@ -313,20 +391,36 @@ sub column_info {
 	SCOPE_CAT SCOPE_SCHEM SCOPE_NAME MAX_CARDINALITY
 	DTD_IDENTIFIER IS_SELF_REF
 	mysql_is_pri_key mysql_type_name mysql_values
+        mysql_is_auto_increment
     );
     my %col_info;
 
     local $dbh->{FetchHashKeyName} = 'NAME_lc';
-    my $desc_sth = $dbh->prepare("DESCRIBE $table_id");
+    # only ignore ER_NO_SUCH_TABLE in internal_execute if issued from here
+    print "DESCRIBE $table_id " . $dbh->quote($column) . "\n";
+    my $desc_sth = $dbh->prepare("DESCRIBE $table_id " . $dbh->quote($column));
     my $desc = $dbh->selectall_arrayref($desc_sth, { Columns=>{} });
 
-    return $desc_sth if $desc_sth->err();
+    #return $desc_sth if $desc_sth->err();
+    if (my $err = $desc_sth->err())
+    {
+        # return the error, unless it is due to the table not 
+        # existing per DBI spec
+        if ($err != $ER_NO_SUCH_TABLE)
+        {
+          return undef;
+        }
+        $dbh->set_err(undef,undef);
+        $desc = [];
+    }
 
     my $ordinal_pos = 0;
     foreach my $row (@$desc) {
 	my $type = $row->{type};
 	$type =~ m/^(\w+)(?:\((.*?)\))?\s*(.*)/;
 	my $basetype = lc($1);
+        my $typemod = $2;
+        my $attr = $3;
 
 	my $info = $col_info{ $row->{field} } = {
 	    TABLE_CAT   => $catalog,
@@ -340,15 +434,17 @@ sub column_info {
 	    ORDINAL_POSITION => ++$ordinal_pos,
 	    mysql_is_pri_key => ($row->{key}  eq 'PRI'),
 	    mysql_type_name  => $row->{type},
+            mysql_is_auto_increment => ($row->{extra} =~ /auto_increment/i ? 1 :
+                                        0),
 	};
 	# This code won't deal with a pathalogical case where a value
 	# contains a single quote followed by a comma, and doesn't unescape
 	# any escaped values. But who would use those in an enum or set?
-	my @type_params = ($2 && index($2,"'")>=0)
-			? ("$2," =~ /'(.*?)',/g)  # assume all are quoted
-			: split /,/, $2||'';      # no quotes, plain list
+	my @type_params = ($typemod && index($typemod,"'")>=0)
+			? ("$typemod," =~ /'(.*?)',/g)  # assume all are quoted
+			: split /,/, $typemod||'';      # no quotes, plain list
 	s/''/'/g for @type_params;                # undo doubling of quotes
-	my @type_attr = split / /, $3||'';
+	my @type_attr = split / /, $attr||'';
 	#warn "$type: $basetype [@type_params] [@type_attr]\n";
 
 	$info->{DATA_TYPE} = SQL_VARCHAR();
@@ -473,6 +569,96 @@ sub primary_key_info {
 }
 
 
+sub foreign_key_info {
+    my ($dbh,
+        $pk_catalog, $pk_schema, $pk_table,
+        $fk_catalog, $fk_schema, $fk_table,
+       ) = @_;
+
+    # INFORMATION_SCHEMA.KEY_COLUMN_USAGE was added in 5.0.6
+    my ($maj, $min, $point) = _version($dbh);
+    return if $maj < 5 || ($maj == 5 && $point < 6);
+
+    my $sql = <<'EOF';
+SELECT NULL AS PKTABLE_CAT,
+       A.REFERENCED_TABLE_SCHEMA AS PKTABLE_SCHEM,
+       A.REFERENCED_TABLE_NAME AS PKTABLE_NAME,
+       A.REFERENCED_COLUMN_NAME AS PKCOLUMN_NAME,
+       A.TABLE_CATALOG AS FKTABLE_CAT,
+       A.TABLE_SCHEMA AS FKTABLE_SCHEM,
+       A.TABLE_NAME AS FKTABLE_NAME,
+       A.COLUMN_NAME AS FKCOLUMN_NAME,
+       A.ORDINAL_POSITION AS KEY_SEQ,
+       NULL AS UPDATE_RULE,
+       NULL AS DELETE_RULE,
+       A.CONSTRAINT_NAME AS FK_NAME,
+       NULL AS PK_NAME,
+       NULL AS DEFERABILITY,
+       NULL AS UNIQUE_OR_PRIMARY
+  FROM INFORMATION_SCHEMA.KEY_COLUMN_USAGE A,
+       INFORMATION_SCHEMA.TABLE_CONSTRAINTS B
+ WHERE A.TABLE_SCHEMA = B.TABLE_SCHEMA AND A.TABLE_NAME = B.TABLE_NAME
+   AND A.CONSTRAINT_NAME = B.CONSTRAINT_NAME AND B.CONSTRAINT_TYPE IS NOT NULL
+EOF
+
+    my @where;
+    my @bind;
+
+    # catalogs are not yet supported by MySQL
+
+#    if (defined $pk_catalog) {
+#        push @where, 'A.REFERENCED_TABLE_CATALOG = ?';
+#        push @bind, $pk_catalog;
+#    }
+
+    if (defined $pk_schema) {
+        push @where, 'A.REFERENCED_TABLE_SCHEMA = ?';
+        push @bind, $pk_schema;
+    }
+
+    if (defined $pk_table) {
+        push @where, 'A.REFERENCED_TABLE_NAME = ?';
+        push @bind, $pk_table;
+    }
+
+#    if (defined $fk_catalog) {
+#        push @where, 'A.TABLE_CATALOG = ?';
+#        push @bind,  $fk_schema;
+#    }
+
+    if (defined $fk_schema) {
+        push @where, 'A.TABLE_SCHEMA = ?';
+        push @bind,  $fk_schema;
+    }
+
+    if (defined $fk_table) {
+        push @where, 'A.TABLE_NAME = ?';
+        push @bind,  $fk_table;
+    }
+
+    if (@where) {
+        $sql .= ' AND ';
+        $sql .= join ' AND ', @where;
+    }
+    $sql .= " ORDER BY A.TABLE_SCHEMA, A.TABLE_NAME, A.ORDINAL_POSITION";
+
+    local $dbh->{FetchHashKeyName} = 'NAME_uc';
+    my $sth = $dbh->prepare($sql);
+    $sth->execute(@bind);
+
+    return $sth;
+}
+
+
+sub _version {
+    my $dbh = shift;
+
+    return
+        $dbh->get_info($DBI::Const::GetInfoType::GetInfoType{SQL_DBMS_VER})
+            =~ /(\d+)\.(\d+)\.(\d+)/;
+}
+
+
 ####################
 # get_info()
 # Generated by DBI::DBD::Metadata
